{% extends "!layout.html" %}

{%- block extrahead %}
{{ super() }}
<script type="text/javascript">
  var _gaq = _gaq || [];
  _gaq.push(['_setAccount', 'UA-15144370-3']);
  _gaq.push(['_trackPageview']);
</script>
{% if not embedded %}<script type="text/javascript" src="{{ pathto('_static/copybutton.js', 1) }}"></script>{% endif %}
{% endblock %}

{% block rootrellink %}
    <li><a href="http://www.github.com/deap/deap"><img src="{{ pathto('_static/deap_icon-39x55.png', 1) }}" style="height: 30px; vertical-align: middle; padding-right: 1em;" />Project Homepage</a>{{ reldelim1 }}</li>
	<li><a href="{{ pathto('index') }}">{{ shorttitle }}</a>{{ reldelim1 }} </li>
{% endblock %}


{% block footer %}
<div class="footer">
    &copy; Copyright {{ copyright|e }}.
    <br />
    Built on {{ last_updated|e }}.
    <a href="{{ pathto('bugs') }}">Found a bug</a>?
    <br />
    Created using <a href="http://sphinx.pocoo.org/">Sphinx</a> {{ sphinx_version|e }}.
</div>
<script type="text/javascript">
  (function() {
     var ga = document.createElement('script');
         ga.src = ('https:' == document.location.protocol ?
                'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
    ga.setAttribute('async', 'true');
    document.documentElement.firstChild.appendChild(ga);
  })();
</script>
{% endblock %}
